/** Bernard Fugas E1-ISIII
	
	Przykad wykorzystania szablonw generate oraz generate_n

	W skrcie:
	Szablony generate i generate_n pozwalaj przypisa wartoci zwracane
	przez bezargumentow funkcj lub obiekt funkcyjny do tablicy
	lub kontenera udostpniajcego iterator.
*/

#include <ostream>
#include <algorithm>	// tu jest definicja generate i generate_n
#include <vector>
#include "boost/random.hpp" //generator liczb losowych - uyty w przykadzie

using namespace std;

/** GeneticAlgorithm to klasa realizujca przykadowy algorytm genetyczny.
	Nie wnikamy w jego implementacj.

	Istotne jest to, e populacja pocztkowa skada si nie tylko z osobnikw
	wygenerowanych losowo, ale rwnie ze znanych "w miar dobrych" rozwiza
	zwanych "ekspertami".
	Szablony generate oraz generate_n pozwalaj w wygodny sposb stworzy
	populacj pocztkow o zadanej procentowej iloci ekspertw.

	Przykady uycia w funkcjach getRandom() oraz createGeneration0()
*/
class GeneticAlgorithm {
public:
	/** Przykadowa klasa osobnika uywanego w algorytmie genetycznym */
	class Individual {
	public:
		static const size_t CHROMOSOME_SIZE = 16;
		Individual() {}
	private:
		unsigned char chromosome_[CHROMOSOME_SIZE];
		friend class GeneticAlgorithm;
		friend ostream& operator<<(ostream& os, const Individual& i);
	};

	typedef vector<Individual> Individuals;
	typedef vector<Individual>::iterator IndividualsIt;
	typedef vector<Individual>::const_iterator IndividualsCit;

	/**
	@brief Konstruktor algorytmu
	@param[in] population_size Rozmiar populacji
	@param[in] expert_percent  Procent ekspertw w populacji pocztkowej
	*/
	GeneticAlgorithm(size_t population_size, size_t expert_percent) :
		population_size_(population_size), 
		expert_percent_(expert_percent)
		{}

	/** 
	@brief Uruchamia szkielet algorytmu.
	*/
	void run(void) {
		bool finish_condition = false;

		createGeneration0(); //stwrz populacj pocztkow

		while( !finish_condition ) {
			//korzystaj z operatorw genetycznych
			//selekcja, krzyowanie, mutacja, ...

			//w w ktrej z iteracji...
			finish_condition = true;
		}
	}

private:
	/**
	@brief Zwraca losowego osobnika
	@return Nowy osobnik o losowym chromosomie
	@note Przykad uycia generate_n
	*/
	static Individual getRandom(void) {
		Individual i;

		/**************************************************************************************
		generate_n(_OutIt _Dest, _Diff _Count, _Fn0 _Func)
		generate_n(_OutElem (&_Dest)[_Size], _Diff _Count, _Fn0 _Func)

		Pozwala przypisa kolejne wartoci funkcji _Func do _Count elementw 
		poczwszy od iteratora lub wskanika _Dest
		**************************************************************************************/
		
		generate_n(i.chromosome_, Individual::CHROMOSOME_SIZE, randomByte);

		return i;
	}

	/** 
	@brief Generuje liczb pseudolosow 
	@return Liczba pseudolosowa z zakresu <0, 255>
	*/
	static unsigned char randomByte(void) {
		static boost::mt19937 state_;
		static boost::uniform_int<> rng(0, 255);
		return static_cast<unsigned char>(rng(state_));
	}

	/**
	@brief Zwraca osobnika "eksperta" - jakie znane, do dobre rozwizanie
	@return Nowy osobnik "ekspert"
	*/
	static Individual getExpert(void) {
		Individual i;
		memcpy(i.chromosome_, GeneticAlgorithm::EXPERT, Individual::CHROMOSOME_SIZE);
		return i;
	}

	/**
	@brief Tworzy populacj pocztkow
	@note Przykad uycia generate
	*/
	void createGeneration0(void) {
		individuals_ = Individuals(population_size_);
		size_t experts_num = population_size_*expert_percent_/100; //liczba ekspertw
		IndividualsIt expertsEnd = individuals_.begin() + experts_num; //iterator "koca ekspertw"


		/**************************************************************************************
		void generate(_FwdIt _First, _FwdIt _Last, _Fn0 _Func)

		Pozwala przypisa elementom od iteratora _First (wcznie z _First) 
		do _Last (bez _Last) kolejne wartoci funkcji _Func
		**************************************************************************************/

		generate(individuals_.begin(), expertsEnd, getExpert); //generuj ekspertw
		generate(expertsEnd, individuals_.end(), getRandom); //generuj losowe osobniki
		
		//mona sobie wyobrazi jeszcze inne rodzaje tworzonych w ten sposb osobnikw...
	}

	static const char EXPERT[Individual::CHROMOSOME_SIZE]; /**< jaki znany, dobry osobnik */
	Individuals individuals_; /**< wektor osobnikw w populacji */
	size_t population_size_;  /**< rozmiar populacji */
	size_t expert_percent_;   /**< procentowy udzia ekspertw w populacji pocztkowej */
	friend ostream& operator<<(ostream& os, const GeneticAlgorithm& ga);
	friend class Individual;
}; //class GeneticAlgorithm

//definicja eksperta
const char GeneticAlgorithm::EXPERT[GeneticAlgorithm::Individual::CHROMOSOME_SIZE] = {
	0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07, 0x08, 0x09, 0x0A, 0x0B, 0x0C, 0x0D, 0x0E, 0x0F, 0x00, };

/**
@brief Zaimplementowany w celu atwego wywietlenia rezultatu
*/
ostream& operator<<(ostream& os, const GeneticAlgorithm::Individual& individual) {
	for(size_t i=0; i<GeneticAlgorithm::Individual::CHROMOSOME_SIZE; ++i) {
		os.setf(ios::hex, ios::basefield);
		os.width(2);
		os.fill('0');
		os << static_cast<unsigned>(individual.chromosome_[i]);
		os.unsetf(ios::hex);
		os << ' ';
	}
	return os;
}

/**
@brief Zaimplementowany w celu atwego wywietlenia rezultatu
*/
ostream& operator<<(ostream& os, const GeneticAlgorithm& ga) {
	for(GeneticAlgorithm::IndividualsCit it=ga.individuals_.begin(); it != ga.individuals_.end(); ++it) 
		os << *it << endl;
	return os;
}

// ================== main ================ 
int main(/*int argc, char** argv*/) {
	size_t population_size = 20; /** < cakowity rozmiar populacji */
	size_t expert_percent = 10; /** < procentowa ilo ekspertw w populacji pocztkowej */
	
	GeneticAlgorithm ga(population_size, expert_percent); //stwrz algorytm
	ga.run();											  //i go uruchom
	//...
	cout << ga;	//wywietl populacj
};
//EOF
